// Etcd client fake for testing purposes. It tries to follow the same
// semantics as the actual etcd (highligh on "tries to"), with a
// simple in-memory implementation.
//
package etcdtest

import (
	"fmt"
	"math/rand"
	"path/filepath"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/autoradio/Godeps/_workspace/src/github.com/coreos/go-etcd/etcd"
)

type datum struct {
	value  string
	expire time.Time
	index  uint64
}

type FakeEtcdClient struct {
	latency time.Duration
	lock    sync.Mutex
	data    map[string]datum
	wlock   sync.Mutex
	watches map[string][]chan *etcd.Response
	index   uint64
}

func NewClient() *FakeEtcdClient {
	return &FakeEtcdClient{
		data:    make(map[string]datum),
		watches: make(map[string][]chan *etcd.Response),
		index:   1,
	}
}

func NewClientWithLatency(maxLatency time.Duration) *FakeEtcdClient {
	return &FakeEtcdClient{
		data:    make(map[string]datum),
		watches: make(map[string][]chan *etcd.Response),
		index:   1,
		latency: maxLatency,
	}
}

func (s *FakeEtcdClient) delay() {
	if s.latency > 0 {
		time.Sleep(time.Duration(rand.Int63n(int64(s.latency))))
	}
}

func (s *FakeEtcdClient) trigger(action, key string) *etcd.Response {
	// Create the etcd Response, containing a single Node.
	resp := &etcd.Response{
		Action:    action,
		EtcdIndex: s.index,
		Node: &etcd.Node{
			Key:           key,
			ModifiedIndex: s.index,
		},
	}
	if action != "delete" {
		resp.Node.Value = s.data[key].value
	}
	s.index++

	// Send notifications to the watchers. To avoid difficult race
	// conditions, we close the channel after having sent a response.
	var toclear []string
	for pfx, w := range s.watches {
		if strings.HasPrefix(key, pfx) {
			for _, ch := range w {
				ch <- resp
				close(ch)
			}
			toclear = append(toclear, pfx)
		}
	}
	for _, pfx := range toclear {
		delete(s.watches, pfx)
	}

	//s.wlock.Unlock()
	s.lock.Unlock()

	return resp
}

func ttlToTime(ttl uint64) time.Time {
	if ttl == 0 {
		return time.Time{}
	}
	return time.Now().Add(time.Duration(ttl) * time.Second)
}

func (s *FakeEtcdClient) AddChild(key, value string, ttl uint64) (*etcd.Response, error) {
	uniqueId := fmt.Sprintf("%x-%x", time.Now().Unix(), rand.Int63())
	return s.Create(filepath.Join(key, uniqueId), value, ttl)
}

func (s *FakeEtcdClient) Create(key, value string, ttl uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	if _, ok := s.data[key]; ok {
		s.lock.Unlock()
		return nil, &etcd.EtcdError{Message: "already there", Index: s.index}
	}

	s.data[key] = datum{
		value:  value,
		expire: ttlToTime(ttl),
		index:  s.index,
	}
	return s.trigger("create", key), nil
}

func (s *FakeEtcdClient) CompareAndDelete(key, oldvalue string, index uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	d, ok := s.data[key]
	if ok && ((oldvalue == "" || d.value == oldvalue) ||
		(index == 0 || d.index != index)) {
		delete(s.data, key)
		return s.trigger("compareAndDelete", key), nil
	}
	s.lock.Unlock()
	return nil, &etcd.EtcdError{Message: "failed", Index: s.index}
}

func (s *FakeEtcdClient) CompareAndSwap(key, value string, ttl uint64, oldvalue string, index uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	d, ok := s.data[key]
	if ok && ((oldvalue == "" || d.value == oldvalue) ||
		(index == 0 || d.index != index)) {
		s.data[key] = datum{
			value:  value,
			expire: ttlToTime(ttl),
			index:  s.index,
		}
		return s.trigger("compareAndSwap", key), nil
	}
	s.lock.Unlock()
	return nil, &etcd.EtcdError{Message: "failed", Index: s.index}
}

func (s *FakeEtcdClient) Delete(key string, recursive bool) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	delete(s.data, key)
	return s.trigger("delete", key), nil
}

func (s *FakeEtcdClient) Get(key string, recursive, boh bool) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()
	now := time.Now()
	resp := &etcd.Response{
		EtcdIndex: s.index,
	}
	var nodes []*etcd.Node
	var modifiedIndex uint64
	key = strings.TrimSuffix(key, "/")
	keyDirPfx := key + "/"
	for path, datum := range s.data {
		if (path == key || strings.HasPrefix(path, keyDirPfx)) &&
			(datum.expire.IsZero() || datum.expire.After(now)) {
			nodes = append(nodes, &etcd.Node{
				Key:           path,
				Value:         datum.value,
				ModifiedIndex: datum.index,
			})
			if datum.index > modifiedIndex {
				modifiedIndex = datum.index
			}
		}
	}
	switch {
	case len(nodes) == 0:
		return nil, &etcd.EtcdError{Message: "not found", Index: s.index}
	case len(nodes) == 1 && nodes[0].Key == key:
		resp.Node = nodes[0]
	default:
		resp.Node = &etcd.Node{
			Key:           key,
			Dir:           true,
			Nodes:         nodes,
			ModifiedIndex: modifiedIndex,
		}
	}
	return resp, nil
}

func (s *FakeEtcdClient) Set(key, value string, ttl uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	s.data[key] = datum{
		value:  value,
		expire: ttlToTime(ttl),
		index:  s.index,
	}
	return s.trigger("set", key), nil
}

func (s *FakeEtcdClient) CreateDir(key string, ttl uint64) (*etcd.Response, error) {
	// TODO. There are no directories.
	return &etcd.Response{}, nil
}

func (s *FakeEtcdClient) Update(key, value string, ttl uint64) (*etcd.Response, error) {
	s.lock.Lock()
	if _, ok := s.data[key]; !ok {
		s.lock.Unlock()
		return nil, &etcd.EtcdError{Message: fmt.Sprintf("key %s not found", key), Index: s.index}
	}
	s.data[key] = datum{
		value:  value,
		expire: ttlToTime(ttl),
		index:  s.index,
	}
	return s.trigger("update", key), nil
}

func (s *FakeEtcdClient) Watch(key string, index uint64, recursive bool, respch chan *etcd.Response, stop chan bool) (*etcd.Response, error) {
	// First scan the data to check for changes >= index.
	s.lock.Lock()
	var nodes []*etcd.Node
	var modifiedIndex uint64
	for k, d := range s.data {
		if strings.HasPrefix(k, key) && d.index >= index {
			nodes = append(nodes, &etcd.Node{
				Key:           k,
				Value:         d.value,
				ModifiedIndex: d.index,
			})
			if d.index > modifiedIndex {
				modifiedIndex = d.index
			}
		}
	}
	if len(nodes) > 0 {
		// Bleargh, we create a fake 'dir' with all contents.
		resp := &etcd.Response{
			Action: "update",
			Node: &etcd.Node{
				Key:           key,
				Dir:           true,
				Nodes:         nodes,
				ModifiedIndex: modifiedIndex,
			},
			EtcdIndex: s.index,
		}
		s.lock.Unlock()
		if respch != nil {
			respch <- resp
		}
		return resp, nil
	}

	// Install the watch.
	ch := respch
	if ch == nil {
		ch = make(chan *etcd.Response, 1)
	}
	s.watches[key] = append(s.watches[key], ch)
	s.lock.Unlock()

	// Wait for a single response.
	var resp *etcd.Response
	if respch != nil {
		<-stop
	} else {
		select {
		case resp = <-ch:
		case <-stop:
		}
	}

	if resp == nil {
		return nil, etcd.ErrWatchStoppedByUser
	}
	return resp, nil
}

func (s *FakeEtcdClient) GetCluster() []string {
	return []string{"http://localhost:2379"}
}

func (s *FakeEtcdClient) SyncCluster() bool {
	return false
}
