package fe

import (
	"fmt"
	"log"
	"math/rand"
	"net"
	"net/url"
	"strconv"
	"strings"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/Godeps/_workspace/src/github.com/miekg/dns"
	"git.autistici.org/ale/autoradio/instrumentation"
)

var (
	// Max number of results for an A query.
	maxResults = 3

	dnsQueryStats  = instrumentation.NewCounter("dns.status")
	dnsTargetStats = instrumentation.NewCounter("dns.target")
)

type etcdAddr struct {
	host string
	port int
	ips  []net.IP
}

// Parsed etcd cluster state.
type etcdClusterState struct {
	addrs []etcdAddr
	ssl   bool
}

func (s *etcdClusterState) IPs() []net.IP {
	var out []net.IP
	for _, addr := range s.addrs {
		out = append(out, addr.ips...)
	}
	return out
}

func parseEtcdClusterState(urls []string) *etcdClusterState {
	if len(urls) == 0 {
		return nil
	}
	var state etcdClusterState
	for _, u := range urls {
		parsedURL, err := url.Parse(u)
		if err != nil {
			continue
		}
		host, portStr, err := net.SplitHostPort(parsedURL.Host)
		if err != nil {
			host = parsedURL.Host
			portStr = "2379"
		}
		port, err := strconv.Atoi(portStr)
		if err != nil {
			continue
		}
		if parsedURL.Scheme == "https" {
			state.ssl = true
		}

		// Resolve the hostname if necessary.
		var ips []net.IP
		if ip := net.ParseIP(host); ip != nil {
			ips = []net.IP{ip}
		} else if resolved, err := net.LookupIP(host); err == nil {
			ips = resolved
		} else {
			continue
		}

		state.addrs = append(state.addrs, etcdAddr{host: dns.Fqdn(host), port: port, ips: ips})
	}
	return &state
}

func addDotToList(l []string) []string {
	var out []string
	for _, s := range l {
		if !strings.HasSuffix(s, ".") {
			s += "."
		}
		out = append(out, s)
	}
	return out
}

// DNSRedirector sends clients to backends using DNS.
//
// The DNSRedirector needs some basic information in order to generate
// the required structural records for the zone (such as NS records).
// Right now the list of nameservers has to be specified manually, but
// it would be nice if we were able to autodetect them, perhaps by
// using the list of nodes (which would require a way to translate
// short name -> fqdn for each node, which is difficult without
// additional requirements on the setup).
type DNSRedirector struct {
	client         *autoradio.Client
	origin         string
	originNumParts int
	publicIps      []net.IP
	etcdCluster    *etcdClusterState
	ttl            int
	soa            dns.RR
	nameservers    []string
	queryTable     map[string]ipFunc
}

// NewDNSRedirector returns a DNS server for the given origin and
// publicIp. The A records served will have the specified ttl.
func NewDNSRedirector(client *autoradio.Client, origin string, publicIps []net.IP, ttl int, nameservers []string) *DNSRedirector {
	if !strings.HasSuffix(origin, ".") {
		origin += "."
	}

	// Create a SOA record for the zone. Some entries will be bogus.
	soa := &dns.SOA{
		Hdr: dns.RR_Header{
			Name:   origin,
			Rrtype: dns.TypeSOA,
			Class:  dns.ClassINET,
			Ttl:    43200,
		},
		Ns:      "ns." + origin,
		Mbox:    "hostmaster." + origin,
		Serial:  uint32(time.Now().Unix()),
		Refresh: 43200,
		Retry:   3600,
		Expire:  uint32(ttl),
		Minttl:  uint32(ttl),
	}

	return &DNSRedirector{
		client:         client,
		origin:         origin,
		originNumParts: len(dns.SplitDomainName(origin)),
		publicIps:      publicIps,
		ttl:            ttl,
		nameservers:    addDotToList(nameservers),
		soa:            soa,

		// We expect the public URLs to just use the origin,
		// but serve a number of extra domain names such as
		// 'stream' and 'www' just in case (and for backwards
		// compatibility reasons).
		queryTable: map[string]ipFunc{
			"":       getAutoradioIPs,
			"www":    getAutoradioIPs,
			"stream": getAutoradioIPs,
			"etcd":   getEtcdIPs,
		},
	}
}

// Periodically update the list of etcd nodes. We need to parse the
// etcd URLs and resolve to a list of IPs.
func (d *DNSRedirector) updateEtcdCluster() {
	d.etcdCluster = parseEtcdClusterState(d.client.EtcdClient.GetCluster())
	go func() {
		for range time.Tick(60 * time.Second) {
			if !d.client.EtcdClient.SyncCluster() {
				continue
			}
			if s := parseEtcdClusterState(d.client.EtcdClient.GetCluster()); s != nil {
				d.etcdCluster = s
			}
		}
	}()
}

// Randomly shuffle a list of addresses.
func shuffle(list []net.IP) []net.IP {
	out := make([]net.IP, len(list))
	for dst, src := range rand.Perm(len(list)) {
		out[dst] = list[src]
	}
	return out
}

// Create skeleton edns opt RR from the query and add it to the
// message m.
func ednsFromRequest(req, m *dns.Msg) {
	for _, r := range req.Extra {
		if r.Header().Rrtype == dns.TypeOPT {
			m.SetEdns0(4096, r.(*dns.OPT).Do())
			return
		}
	}
	return
}

func (d *DNSRedirector) withOrigin(name string) string {
	if name == "" {
		return d.origin
	}
	return name + "." + d.origin
}

// Create an A resource record.
func (d *DNSRedirector) newA(name string, ip net.IP) dns.RR {
	return &dns.A{
		Hdr: dns.RR_Header{
			Name:   d.withOrigin(name),
			Rrtype: dns.TypeA,
			Class:  dns.ClassINET,
			Ttl:    uint32(d.ttl),
		},
		A: ip,
	}
}

// Create an AAAA resource record.
func (d *DNSRedirector) newAAAA(name string, ip net.IP) dns.RR {
	return &dns.AAAA{
		Hdr: dns.RR_Header{
			Name:   d.withOrigin(name),
			Rrtype: dns.TypeAAAA,
			Class:  dns.ClassINET,
			Ttl:    uint32(d.ttl),
		},
		AAAA: ip,
	}
}

// Strip the origin from the query.
func (d *DNSRedirector) getQuestionName(q dns.Question) string {
	lx := dns.SplitDomainName(q.Name)
	ql := lx[0 : len(lx)-d.originNumParts]
	return strings.ToLower(strings.Join(ql, "."))
}

// Flatten IPs from the list of nodes.
func flattenIPs(nodes []*autoradio.NodeStatus) []net.IP {
	var ips []net.IP
	for _, n := range nodes {
		ips = append(ips, n.IP...)
	}
	return ips
}

type ipFunc func(*DNSRedirector, dns.Question) []net.IP

func getEtcdIPs(d *DNSRedirector, q dns.Question) []net.IP {
	if d.etcdCluster == nil {
		return nil
	}

	// Always serve all IPs.
	return filterIPByProto(d.etcdCluster.IPs(), q.Qtype == dns.TypeAAAA)
}

func getAutoradioIPs(d *DNSRedirector, q dns.Question) []net.IP {
	// Serve all active nodes on every request. We don't really
	// care about errors from GetNodes as long as some nodes are
	// returned (i.e. stale data from the cache is accepted).
	// Also, we need to filter the resulting list for nodes whose
	// IP address protocol version matches the request type (IPv4
	// for A requests, IPv6 for AAAA).
	nodes, err := d.client.GetNodes()
	if err != nil {
		log.Printf("error fetching list of nodes: %v", err)
		return nil
	}

	// Shuffle the list in random order, and keep only the first N
	// results.
	ips := shuffle(filterIPByProto(flattenIPs(nodes), q.Qtype == dns.TypeAAAA))
	if len(ips) > maxResults {
		ips = ips[:maxResults]
	}
	return ips
}

func (d *DNSRedirector) handleQuestion(q dns.Question, m *dns.Msg) bool {
	query := d.getQuestionName(q)
	switch {
	case query == "" && q.Qtype == dns.TypeSOA:
		m.Answer = append(m.Answer, d.soa)
		return true

	case query == "" && q.Qtype == dns.TypeNS:
		for _, ns := range d.nameservers {
			m.Answer = append(m.Answer, &dns.NS{
				Hdr: dns.RR_Header{
					Name:   d.withOrigin(query),
					Rrtype: dns.TypeNS,
					Class:  dns.ClassINET,
					Ttl:    3600,
				},
				Ns: ns,
			})
		}
		return true

	case q.Qtype == dns.TypeSRV:
		if d.etcdCluster == nil {
			return false
		}

		// Serve the right name depending on whether the etcd
		// cluster is set up to use SSL or not.
		if !((query == "_etcd-server-ssl._tcp" && d.etcdCluster.ssl) || (query == "_etcd-server._tcp" && !d.etcdCluster.ssl)) {
			return false
		}

		for _, addr := range d.etcdCluster.addrs {
			rec := &dns.SRV{
				Hdr: dns.RR_Header{
					Name:   d.withOrigin(query),
					Rrtype: dns.TypeSRV,
					Class:  dns.ClassINET,
					Ttl:    uint32(d.ttl),
				},
				Target: addr.host,
				Port:   uint16(addr.port),
			}
			m.Answer = append(m.Answer, rec)
		}
		return true

	case q.Qtype == dns.TypeA || q.Qtype == dns.TypeAAAA:
		// Look up the requested name in the routing table,
		// and retrieve the list of IPs for the reply. Return
		// an NXDOMAIN for unknown queries.
		ipfn, ok := d.queryTable[query]
		if !ok {
			return false
		}

		// Get the IPs for the reply.
		ips := ipfn(d, q)
		if len(ips) == 0 {
			// In case of errors, if the IP list is empty, fall
			// back to serving our public IP (just to avoid
			// returning an empty reply, which might be cached for
			// longer).
			log.Printf("fallback to localhost")
			ips = d.publicIps
		}

		for _, ip := range ips {
			var rec dns.RR
			if q.Qtype == dns.TypeAAAA {
				rec = d.newAAAA(query, ip)
			} else {
				rec = d.newA(query, ip)
			}
			m.Answer = append(m.Answer, rec)
			dnsTargetStats.IncrVar(ipToMetric(ip))
		}
		return true
	}
	return false
}

func responseToString(m *dns.Msg) string {
	if m.MsgHdr.Rcode != dns.RcodeSuccess {
		return dns.RcodeToString[m.MsgHdr.Rcode]
	}
	var out []string
	for _, ans := range m.Answer {
		var s string
		switch t := ans.(type) {
		case *dns.A:
			s = fmt.Sprintf("A %s", t.A)
		case *dns.AAAA:
			s = fmt.Sprintf("AAAA %s", t.AAAA)
		case *dns.SRV:
			s = fmt.Sprintf("SRV %s:%d", strings.TrimRight(t.Target, "."), t.Port)
		default:
			s = t.String()
		}
		out = append(out, s)
	}
	return strings.Join(out, ", ")
}

func (d *DNSRedirector) ServeDNS(w dns.ResponseWriter, req *dns.Msg) {
	m := new(dns.Msg)

	// Only consider the first question.
	q := req.Question[0]

	// Just NACK ANYs
	if q.Qtype == dns.TypeANY {
		m.SetRcode(req, dns.RcodeServerFailure)
		ednsFromRequest(req, m)
		w.WriteMsg(m)
		return
	}

	// Handle the request.
	ednsFromRequest(req, m)
	if !d.handleQuestion(q, m) {
		m.SetRcode(req, dns.RcodeNameError)
		dnsQueryStats.IncrVar("error")
	} else {
		m.SetReply(req)
		m.MsgHdr.Authoritative = true
		dnsQueryStats.IncrVar("ok")
	}

	log.Printf("[%d] %s %s %s -> %s", req.MsgHdr.Id, w.RemoteAddr(), q.Name, dns.TypeToString[q.Qtype], responseToString(m))

	w.WriteMsg(m)
}

// Start the DNS servers on the given address (both tcp and udp).
// It creates new goroutines and returns immediately.
func (d *DNSRedirector) Start(addr string) {
	d.updateEtcdCluster()

	dns.Handle(d.origin, d)

	for _, proto := range []string{"tcp", "udp"} {
		go func(proto string) {
			server := &dns.Server{Addr: addr, Net: proto}
			log.Printf("Starting DNS server on %s/%s", addr, proto)
			log.Fatal(server.ListenAndServe())
		}(proto)
	}
}
