package node

import (
	"encoding/xml"
	"errors"
	"flag"
	"fmt"
	"io"
	"log"
	"net"
	"net/http"
	"os"
	"os/exec"
	"strconv"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
	"git.autistici.org/ale/autoradio/util"
)

// Managing the local Icecast instance is a less simple task than it
// might look, as the details tend to be distribution-specific. We'd
// like to avoid implementing yet another service manager within
// radiod, so we are just going to assume that the Icecast daemon is
// managed using the local distribution-specific tools, and is running
// independently of radiod. Furthermore, we're going to make the
// following assumptions:
//
// - We have permissions to modify the Icecast config file;
//
// - We have installed our status XSLT page;
//
// - We can reload the Icecast service (send the icecast2 process a
// SIGHUP). Ideally we'd like to use the distribution-specific method,
// but with systemd we need to use "sudo". The code tries to
// autodetect a supported mechanism.
//
// The above is usually accomplished by running radiod as the same
// user that is running the Icecast daemon.
//
var (
	icecastConfigFile      = flag.String("icecast-config", "/etc/icecast2/icecast.xml", "Icecast configuration file")
	icecastCustomReloadCmd = flag.String("icecast-reload-command", "", "Command to reload / restart the icecast2 daemon")

	icecastStatusPage = "/status-autoradio.xsl"
	icecastOk         = instrumentation.NewGauge("icecast.ok")
)

const (
	systemCtlReloadCmd = "sudo systemctl reload icecast2.service || sudo systemctl restart icecast2.service"
	debianReloadCmd    = "/usr/sbin/service icecast2 reload || /usr/sbin/service icecast2 restart"
	genericReloadCmd   = "pkill -HUP icecast2"
)

func getIcecastReloadCmd() string {
	if *icecastCustomReloadCmd != "" {
		return *icecastCustomReloadCmd
	}
	if _, err := os.Stat("/bin/systemctl"); err == nil {
		log.Printf("using /bin/systemctl to reload icecast2")
		return systemCtlReloadCmd
	}
	if _, err := os.Stat("/usr/sbin/service"); err == nil {
		log.Printf("using /usr/sbin/service to reload icecast2")
		return debianReloadCmd
	}
	log.Printf("using pkill to reload icecast2")
	return genericReloadCmd
}

// Icecast returns empty fields in our status handler, which we'll
// need to turn into integers (the xml unmarshaler will return an
// error in this specific case), so we use a separate type for
// decoding the status page output. This would be much simpler if I
// knew how to get the XSLT to put a default value in the output
// instead of an empty field...
type icecastMountStatusRaw struct {
	Name         string `xml:"name,attr"`
	Listeners    string `xml:"listeners"`
	BitRate      string `xml:"bitrate"`
	Quality      string `xml:"quality"`
	VideoQuality string `xml:"video-quality"`
	FrameSize    string `xml:"frame-size"`
	FrameRate    string `xml:"frame-rate"`
}

type icecastStatusRaw struct {
	XMLName xml.Name                `xml:"status"`
	Mounts  []icecastMountStatusRaw `xml:"mount"`
}

type icecastStatus struct {
	Mounts []autoradio.IcecastMountStatus
	Up     bool
}

type icecastController struct {
	config    *icecastConfig
	status    *icecastStatus
	reloadCmd string
	stop      chan bool
}

func newIcecastController(publicIP string, maxClients int) *icecastController {
	return &icecastController{
		config:    newIcecastConfig(publicIP, maxClients),
		status:    &icecastStatus{},
		reloadCmd: getIcecastReloadCmd(),
	}
}

// Reload the icecast daemon. Redirects output to our standard error
// for debugging purposes.
func (ic *icecastController) reload() error {
	log.Printf("reloading icecast")
	cmd := exec.Command("/bin/sh", "-c", ic.reloadCmd)
	cmd.Dir = "/"
	cmd.Stdout = os.Stderr
	cmd.Stderr = os.Stderr
	return cmd.Run()
}

// Kill sources connected to local streams.
func (ic *icecastController) killSources(conf *clusterConfig) {
	client := &http.Client{}
	for _, m := range conf.ListMounts() {
		req, err := http.NewRequest("GET", fmt.Sprintf("http://localhost:%d/admin/killsource?mount=%s", autoradio.IcecastPort, autoradio.MountNameToIcecastPath(m.Name)), nil)
		if err != nil {
			log.Printf("killSources: %v", err)
			continue
		}
		req.SetBasicAuth("admin", getIcecastAdminPassword())
		resp, err := client.Do(req)
		if err != nil {
			log.Printf("killSources: %v", err)
			continue
		}
		resp.Body.Close()
		// Do not report errors when killing sources on relay
		// mounts - we only do it in the off chance that the
		// mount was *not* a relay before.
		if !m.IsRelay() && resp.StatusCode != 200 {
			log.Printf("killSources: HTTP status %s", resp.Status)
		}
	}
}

// Update reloads the Icecast daemon with a new configuration.
func (ic *icecastController) Update(conf *clusterConfig, isMaster bool, masterAddr net.IP) error {
	if !isMaster && masterAddr == nil {
		return errors.New("unknown system state")
	}

	// Write a new configuration (atomically).
	ic.config.Update(conf, isMaster, masterAddr.String())
	data, err := ic.config.Encode()
	if err != nil {
		return err
	}

	changed, err := util.WriteFileIfChanged(*icecastConfigFile, data)
	if err != nil {
		return err
	}
	if changed {
		// Try to kill sources connected to the local icecast daemon
		// before reloading, otherwise we'll have problems on master
		// -> slave transitions (for example, on a restart of radiod)
		// as sources will be "stuck" preventing the new configuration
		// from taking effect.
		//
		// There is unfortunately a small race condition here, as it
		// is possible for sources to reconnect before we reload the
		// icecast daemon.
		//
		// TODO: only kill sources whose configuration has changed.
		ic.killSources(conf)

		// Reload the icecast daemon.
		return ic.reload()
	}

	return nil
}

func (ic *icecastController) GetStatus() *icecastStatus {
	return ic.status
}

func (ic *icecastController) Run(stop chan bool) {
	t := time.NewTicker(3 * time.Second)
	downStatus := &icecastStatus{}
	for {
		select {
		case <-t.C:
			if status, err := ic.fetchStatus(); err == nil {
				ic.status = status
				icecastOk.Set(1)
			} else {
				log.Printf("error checking iceast status: %v", err)
				ic.status = downStatus
				icecastOk.Set(0)
			}
		case <-stop:
			return
		}
	}
}

func (ic *icecastController) fetchStatus() (*icecastStatus, error) {
	resp, err := http.Get(fmt.Sprintf("http://localhost:%d%s", autoradio.IcecastPort, icecastStatusPage))
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	return ic.parseStatusPage(resp.Body)
}

func (ic *icecastController) parseStatusPage(input io.Reader) (*icecastStatus, error) {
	var ustatus icecastStatusRaw
	if err := xml.NewDecoder(input).Decode(&ustatus); err != nil {
		return nil, err
	}

	// Quick converters from string that default to 0.
	toi := func(s string) int {
		if i, err := strconv.Atoi(s); err == nil {
			return i
		}
		return 0
	}
	tof := func(s string) float64 {
		if f, err := strconv.ParseFloat(s, 64); err == nil {
			return f
		}
		return 0
	}

	status := icecastStatus{
		Up:     true,
		Mounts: make([]autoradio.IcecastMountStatus, 0, len(ustatus.Mounts)),
	}
	for _, um := range ustatus.Mounts {
		m := autoradio.IcecastMountStatus{
			Name:         autoradio.IcecastPathToMountName(um.Name),
			Listeners:    toi(um.Listeners),
			BitRate:      toi(um.BitRate),
			Quality:      tof(um.Quality),
			VideoQuality: tof(um.VideoQuality),
			FrameSize:    um.FrameSize,
			FrameRate:    tof(um.FrameRate),
		}
		status.Mounts = append(status.Mounts, m)
	}

	return &status, nil
}
